#!Spriteモジュール
#|テクスチャを管理するモジュール。テクスチャ画像を読み込み、それをセルに切り分けて管理します。

module Sprite
@texhash={} # key=any value=[ Texture, celldata , NMV]
	def dump
		#$Sprite.dump
		#@現在のスプライトの状態をコンソールに出力します。
		p @texhash
	end
	def clear
		#$Sprite.clear
		#@ロードされているスプライトすべてを廃棄します。
		@texhash.each{|item|
			item[1][0].delete
			if item[1][2] then
				item[1][2].delete
			end
		}
		@texhash={}
	end
	def update
		#$Sprite.update
		#@ムービーテクスチャのアップデートをします。Game.updateで呼んでください。
		@texhash.each{|item|
			if item[1][2] then
				item[1][2].totexture(item[1][0])
			end
		}
	end
	def load (key,filename,loopflag=true,hasalpha=false)
		#$Sprite.load (key,String filename[,Boolean loopflag=true,Boolean hasalpha=false])
		#@スプライトをロードします。rubyのハッシュのキーには数値も文字列もシンボルも使えるので、状況で使い分けてください。
		#@これは画像を読み込む関数です。セルの切り分け等は他の関数で行います。デフォルトでは、全体をセル0とみなします。
		#@filenameにはNMVファイルも指定できます。この場合ムービーテクスチャになります。ムービーテクスチャの場合はループするかどうか、α値を持つムービーかどうかを指定できます。
		#@filenameに"*幅,高さ,色"(例："*800,600,0xFFFFFFFF")とすれば塗りつぶし画像を指定できます。

		if @texhash[key] then Sprite.delete(key) end
		if filename[0,1]=="*" then
			p=filename[1,1000].split(",")
			tex=Texture.new(p[0].to_i,p[1].to_i)
			hex=p[2]
			if hex[0,2]=="0x" then hex=hex[2,1000] end
			c=hex[0,2].hex.to_f*0x1000000+hex[2,1000].hex.to_f
			tex.fill(c)
		elsif filename[-3,3].downcase=="nmv" then
			nmv=NMV.new(filename)
			tex=Texture.new(nmv.getwidth,nmv.getheight)
			nmv.play(loopflag,hasalpha)
			nmv.totexturewait(tex)
		else
			tex=Texture.new(filename)
			nmv=nil
		end

		@texhash[key]=[ tex,[[0,0,tex.getwidth,tex.getheight]],nmv]
	end
	def text (key,text,x,y,font)
		#$Sprite.text(key,text,x,y,font)
		#@スプライトの座標x,yにフォントfont(Fontオブジェクトです)で文字列textを描画します。
		if @texhash[key] then
			font.put(@texhash[key][0],text,x,y)
		else
			raise("存在しないテクスチャに描画しようとしました。")
		end
	end
	def setcelldata (key,celldata)
		#$Sprite.setcelldata(key,celldata)
		#@読んだスプライトをセルに切り分ける情報を設定する関数です。
		#@json等で読み込んだものか、rubyスクリプトで記述して作った配列データを割り当ててください。
		#@セルデータは１次元配列で、セル0から順にデータを格納してください。一セル分のデータは、これも配列で、[sx,sy,sw,sh]です。
		#@sx,syはテクスチャ上の左上座標、sw,shはセルの幅と高さです。ピクセル単位です。
		if @texhash[key]==nil then raise("ロードしていないテクスチャを切り分けようとしました。") end
		@texhash[key][1]=celldata
	end
	def slice (key,w,h,space_x=0,space_y=0)
		#$Sprite.slice(key,Fixnum w,Fixnum h[,Fixnum space_x,Fixnum space_y])
		#@読んだスプライトをセルに切り分ける関数です。幅と高さで格子状に切り分けます。
		#@space_xとspace_yに値を指定すると、チップの周囲（上下左右端も含む）に、それだけのスペースが開いているものとみなします。
		#@これは、外部ツールの一部にはテクスチャの拡大縮小補間対策にスペースを空けてあるものがあるからです。
		#@space_xとspace_yを指定するときには、wとhはスペースの分を含まないチップそのものの大きさを指定してください。
		if @texhash[key]==nil then raise("ロードしていないテクスチャを切り分けようとしました。") end
		celldata=[]
		tw=@texhash[key][0].getwidth
		th=@texhash[key][0].getheight
		cx=(tw/(w+space_x)).floor-1
		cy=(th/(h+space_y)).floor-1
		for j in 0..cy do
			for i in 0..cx do
				celldata.push [i*(w+space_x)+space_x,j*(h+space_y)+space_y,w,h]
			end
		end
		@texhash[key][1]=celldata
	end
	def slicenum (key,xnum,ynum)
		#$Sprite.slicenum (key,xnum,ynum)
		#@読んだスプライトをセルに切り分ける関数です。こちらは横方向と縦方向の分割数を指定します。
		if @texhash[key]==nil then raise("ロードしていないテクスチャを切り分けようとしました。") end
		slice(key,(@texhash[key][0].getwidth / xnum).floor,(@texhash[key][0].getheight / ynum).floor)
	end
	def gettexture(key)
		#$Sprite.gettexture(key) → Texture
		#@テクスチャオブジェクトを返します。
		return @texhash[key][0]
	end
	def delete (key)
		#$Sprite.delete(key)
		#@指定されたスプライトを削除します。
		@texhash[key][0].delete
		if texhash[key][2] then texhash[key][2].delete end
		@texhash[key]=nil
	end
	def copy (key_dest,key_src)
		#$Sprite.copy(key_dest,key_src)
		#@key_destで指定されたキーのスプライトを作成し、key_srcで指定されたキーのスプライトをコピーします。
		#@テクスチャを加工したいけど元のものも残しておきたい、というようなときに使えます。
		if @texhash[key_dest] then Sprite.delete(key_dest) end
		if @texhash[key_src][2] then raise("ムービーテクスチャはコピーできません。") end
		@texhash[key_dest]=[@texhash[key_src][0].clone,@texhash[key_src][1].clone,nil]
	end
	def draw(key,cell)
		#$Sprite.draw(key,Fixnum cell)
		#@キーとセルを指定し、描画します。Draw.fromとDraw.putを内部で呼び出します。
		if @texhash[key] then
			t=@texhash[key]
			Draw.from(t[1][cell][0],t[1][cell][1],t[1][cell][2],t[1][cell][3])
			Draw.put(t[0])
		else
			raise("存在しないテクスチャを描画しようとしました。")
		end
	end
	def get (key,cell)
		#$Sprite.get(key,Fixnum cell) → [Texture tex,Fixnum sx,Fixnum sy,Fixnum sw,Fixnum sh]
		#@キーとセルを指定し、Textureインスタンス、転送元x,y,幅、高さの情報が入った配列を得ます。見つからなかった場合はnilが戻ります。
		if @texhash[key] then
			t=@texhash[key]
			if t[1].size<=cell || cell<0 then return nil end
			return [t[0],t[1][cell][0],t[1][cell][1],t[1][cell][2],t[1][cell][3]]
		else
			return nil
		end
	end
	def isplaying (key)
		#$Sprite.isplaying(key)→Boolean
		#@キーで指定されたスプライトがムービーテクスチャである場合、再生中ならtrue、停止していればfalseを返します。
		#@ムービーテクスチャでない場合は常にfalseを返します。
		if @texhash[key] then
			if @texhash[key][2] then
				return @texhash[key][2].isplaying
			else
				return false
			end
		else
			raise("存在しないスプライトの再生状況を取得しようとしました。")
		end
	end
	def monotone(key,color)
		#$Sprite.monotone(key,color)
		#@スプライトを指定した色(0xAARRGGBB)でモノトーンにします。
		if @texhash[key] then
			@texhash[key][0].monotone(color)
		else
			raise("存在しないテクスチャに描画しようとしました。")
		end
	end
	def mask(key)
		#$Sprite.mask(key)
		#@スプライトをアルファチャンネル化（アルファ値だけが残った白色画像）にします。
		if @texhash[key] then
			@texhash[key][0].mask
		else
			raise("存在しないテクスチャに描画しようとしました。")
		end
	end
	module_function :dump,:clear,:update,:load,:slice,:slicenum,:delete,:copy,:get,:draw,:monotone,:mask,:gettexture,:text
end

#!TiledMapクラス
#|Spriteモジュールにロードされたマップチップを利用してタイルマップを表示するクラスです。
#|マップはラップ表示されるので、ラップさせたくない場合は各端でスクロールを止めてキャラクターのほうを動かすなどの処理が必要です。
class TiledMap
	def initialize(spkey)
		#$TiledMap.new(spkey)→TiledMap
		#@タイルマップオブジェクトを作成します。Spriteモジュールのキーで、マップチップを指定します。
		#@マップの１チップの大きさはセル０番の大きさで設定されます。
		@w=nil
		@h=nil
		@lx=nil
		@ly=nil
		@xnum=nil
		@ynum=nil
		@data=nil
		@spkey=spkey
		data=Sprite.get(spkey,0) #とりあえず0番のサイズで測る。
		@cx=data[3]
		@cy=data[4]
		gw=Game.getwidth
		gh=Game.getheight
		@xnum=(gw/@cx).ceil
		@ynum=(gh/@cy).ceil
		@anim={}
		@count=0
	end
	def setanimation(mapanimation_data)
		#$tmap.setanimation(mapanimation_data)
		#@タイルマップのセルアニメーションを設定します。
		#@mapanimation_dataはハッシュです。アニメーションさせたいセル番号をキーとして、値は[アニメセル枚数,一コマあたりのフレーム数]です。
		#@表示セル番号はセル枚数までは1ずつ増えていき、それを超えるとデータ上のセル番号に戻ります。
		#@つまり、{1=>[3,10]}の場合だと、セル番号1のチップは、10フレームごとに1 2 3 1 2 3 1 ... とアニメーションします。
		#@あまり細かい制御は出来ませんが、複雑なアニメーションをさせたい場合はAnimatorを作ってそれを重ねて表示してください。
		if !mapanimation_data.kind_of?(Hash) then raise "setanimationにはハッシュを渡してください。" end
		@anim=mapanimation_data
	end
	def draw (x,y,resolution,alpha=nil) 
		#$tmap.draw(x,y,resolution[,alpha])
		#@マップ上の座標x,yを左上にして画面に描画します。単位はチップサイズ/resolutionです。alphaは指定しない場合255になります。
		#@resolutionは指定した値の解像度を示します。チップサイズを割る数を指定します。つまり、4を指定すればxとyは4で1チップ分になります。
		#@これは、半キャラ・1/4キャラ等の単位でスクロールさせたい時に使います。整数を指定してください。
		lx=-((x%resolution)*@cx/resolution).floor
		ly=-((y%resolution)*@cy/resolution).floor
		mx=(x / resolution).floor
		my=(y / resolution).floor
		for j in -1..@ynum
			for i in -1..@xnum
				cell=@data[((i+mx)%@w)+((j+my)%@h)*@w]
				if cell==-1 then continue end
				Draw.to(i*@cx+lx,j*@cy+ly)
				if alpha then Draw.alpha(alpha) end
				if @anim[cell] then
					num=@anim[cell][0]
					tm=@anim[cell][1]
					cell=cell+((@count % (num*tm))/tm).floor
					Sprite.draw(@spkey,cell)
				else
					Sprite.draw(@spkey,cell)
				end
			end
		end
		@count=@count+1
	end
	def setdata (w,h,data) 
		#$tmap.setdata(w,h,data)
		#@マップデータを設定します。w,hは幅と高さ、dataはチップ番号(=セル番号)の入った１次元配列でサイズがw*hである必要があります。チップ番号-1は表示しないことを示します（重ねあわせ用）
		#@マップデータは配列を参照しているだけです、コピーはしていません。dataを呼び出し側でも取っておけば、それをいじればマップを動的に変えることが出来ます。
		#@なお、アニメーションの同期のためのタイマはタイルマップオブジェクト内で共通です。
		@data=data
		@w=w
		@h=h
		if data.size!=w*h then raise "マップサイズと与えたマップデータのサイズが一致しません。" end
	end
	def get(x,y)
		#$tmap.get(x,y)
		#@指定したマップ座標のチップ番号を得ます。drawと違い、こちらはresolutionが設定できません。
		#@なお、getで得られる値はデータ上のものです。アニメーション上のセル番号は表示するときだけに使われます。
		return data[x+y*@h]
	end
end

#!Animatorクラス
#|セルアニメーションするオブジェクトを管理するクラスです。内部でそれぞれのフレームの時間も管理しています。
class Animator
	def initialize (spkey,animation_list)
		#$Animator.new(spkey,animation_list)
		#@Animatorオブジェクトを作成します。spkeyはアニメーションが使用するスプライトのキーです。
		#@animation_listはハッシュで、アニメーションの種類をキー(例えば:down :upのように)、animation_dataを値に持ちます。
		#@キー0はデフォルトのアニメーションとして扱われます。作成したばかりでsetを指定していない場合、アニメーションanimation_list[0]を指定したことになります。ない場合、drawでエラーになります。デフォルトを作らない場合は、必ずdrawより前にsetでアニメーションキーをセットしてください。
		#@animation_dataは配列で、[[cell,time],[cell,time],...]という形式になっています。それぞれのセル番号と時間（フレーム数単位）です。
		#@timeに-1を指定すると、ループせずそこでアニメーションが止まります。timeに-1が指定されていなかった場合、配列の最後の要素まで終わると先頭に戻ります。
		#@アニメーションがひとつしかない場合は、animation_listハッシュではなくanimation_data配列を渡すことが出来ます。自動でそれがキー0のアニメーションとして設定されます。
		#@アニメーションがひとつで、１セルしかない場合（つまり静止している場合）は、セル番号のみを渡せます。自動でそれがキー0のアニメーション、１セルのみで停止、として設定されます。
		@spkey=spkey
		if animation_list.kind_of?(Fixnum) then
			@list={0=>[[animation_list,-1]]}
		elsif animation_list.kind_of?(Array) then
			@list={0=>animation_list}
		else
			@list=animation_list
		end
		@framecount=0
		@animno=0
		@animkey=0
		@pauseflag=false
	end
	def set (animkey)
		#$anim.set(animkey)
		#@アニメーションのキーを指定してアニメを開始します。
		#@アニメーションのキーが現在と同じ場合は何もしません。リセットもしません。
		if @animkey==animkey then return end
		if @list[animkey]==nil then raise "存在しないアニメーションキーを指定しました。" end
		@animkey=animkey
		@framecount=0
		@animno=0
	end
	def pause
		#$anim.pause
		#@アニメーションを一時停止します。
		@pauseflag=true
	end
	def resume
		#$anim.resume
		#@アニメーションの一時停止を解除します。一時停止していない場合何もしません。
		@pauseflag=false
	end
	def reset
		#$anim.reset
		#@現在のキーのアニメーションを最初から再生します。
		@framecount=0
		@animno=0
	end
	def isplaying
		#$anim.isplaying→Boolean
		#@セルがひとつしかない場合（つまり止まってる場合）、もしくはアニメーションがtime=-1で止まっている場合はfalse、それ以外のアニメーションを再生中ならtrueを戻します。
		if @list[@animkey][@animno][1]==-1 or @list[@animkey].size<=1 then return false else return true end
	end
	def draw (alpha=nil)
		#$anim.draw([alpha])
		#@アニメーション状態に合わせて描画します。alphaは指定しないと255になります。
		celltime=@list[@animkey][@animno][1]
		if celltime==-1 then
			#何もしない
		elsif @framecount>=celltime then
			@framecount=0
			@animno=@animno+1
			if @animno>=@list[@animkey].size then
				@animno=0
			end
		else
			if !@pauseflag then
				@framecount=@framecount+1
			end
		end
		if alpha then Draw.alpha(alpha) end
		Sprite.draw(@spkey,@list[@animkey][@animno][0])
	end
end

#!Windowクラス
#|テキスト等を表示するウィンドウを管理するクラスです。スプライトとして「ウィンドウ枠」を指定する必要があります。
#|ウィンドウ枠はセル0～8がそれぞれ同サイズの正方形だと想定し、左上から右下に対応し、それぞれをタイル上に並べてウィンドウを作ります。
#|（正方形スプライト画像をロードしてslicenumで3,3を掛ければ一番簡単にそのような設定になりますが、他の方法でそうしても構いません。）
#|Windowクラスのインスタンスは内部に描画したテキストを記憶するTextureを持っていますので、使い終わったらdeleteしたほうがリソースの節約になります。
class Window
	def initialize (spkey,w,h)
		#$Window.new(spkey,w,h)→UI
		#@ウィンドウを作成します。spkeyはスプライトのキー、w,hは幅と高さです。単位はスプライトのセルの幅です（ピクセルではありません）。
		#@たとえば10,5を指定すると、スプライト枠のセル4番（真ん中）の画像を10x5個並べて、その周りに適切な枠の画像を並べることになります。
		@sz=Sprite.get(spkey,0)[3]	#枠のセルのサイズ（正方形を仮定）
		@txt=Texture.new(@sz*w+10,@sz*h+10) #テクスト保持用のテクスチャ、少し大きめに作ってある。
		@txt.fill(0)
		@w=w
		@h=h
		@spkey=spkey
		@seq_text=nil
		@seq_no=nil
		@seq_x=nil
		@seq_y=nil
		@seq_f=nil
	end
	def delete
		#$Window.delete
		#@ウィンドウを明示的に削除します。
		@txt.delete
	end
	def draw(x,y,alpha=nil)
		#$wnd.draw(x,y,[alpha])
		#@ウィンドウを指定位置に描画します。alphaの指定を省略すると255になります。
		#@枠を表示する
		Draw.to(x-@sz,y-@sz) #左上
		if alpha then Draw.alpha(alpha) end
		Sprite.draw(@spkey,0)

		Draw.to(x+@w*@sz,y-@sz) #右上
		if alpha then Draw.alpha(alpha) end
		Sprite.draw(@spkey,2)

		Draw.to(x-@sz,y+@h*@sz) #左下
		if alpha then Draw.alpha(alpha) end
		Sprite.draw(@spkey,6)

		Draw.to(x+@w*@sz,y+@h*@sz) #右下
		if alpha then Draw.alpha(alpha) end
		Sprite.draw(@spkey,8)

		for i in 0..(@w-1) do
			Draw.to(x+i*@sz,y-@sz) #上列
			if alpha then Draw.alpha(alpha) end
			Sprite.draw(@spkey,1)
			Draw.to(x+i*@sz,y+@h*@sz) #下列
			if alpha then Draw.alpha(alpha) end
			Sprite.draw(@spkey,7)
		end

		for i in 0..(@h-1) do
			Draw.to(x-@sz,y+i*@sz) #左列
			if alpha then Draw.alpha(alpha) end
			Sprite.draw(@spkey,3)
			Draw.to(x+@w*@sz,y+i*@sz) #右列
			if alpha then Draw.alpha(alpha) end
			Sprite.draw(@spkey,5)
		end

		for j in 0..(@h-1) do
			for i in 0..(@w-1) do
				Draw.to(x+i*@sz,y+j*@sz) #真ん中
				if alpha then Draw.alpha(alpha) end
				Sprite.draw(@spkey,4)
			end
		end

		Draw.to(x,y)
		if alpha then Draw.alpha(alpha) end
		Draw.put(@txt)
	end
	def text(text,x,y,font)
		#$wnd.text(text,x,y,font)
		#@ウィンドウに文字列を書き込みます。x,yは座標です。(枠ではなく文字を表示する部分の左上からの相対座標、ピクセル単位）。
		#@fontはフォント、textは表示する文字列です。戻り値は、書き込んだ文字数です。
		return font.put(@txt,text,x,y)
	end
	def textstart(text,x,y,font)
		#$wnd.textstart(text,x,y,font)
		#@テキストを一文字ずつ表示するモードをスタートします。このメソッドを呼び出しただけでは一文字も書かれません。
		#@呼び出してからnextを呼ぶと、そのたびに一文字ずつ書いていきます。
		@seq_text=Font.split(text)
		@seq_no=0
		@seq_x=x
		@seq_y=y
		@seq_f=font
	end
	def textnext
		#$wnd.textnext→Boolean
		#@textstartで設定した文字を一文字書き込み、それでテキスト書き込みが終了したらtrue、まだならばfalseを返します。
		if !@seq_text then raise "テキストが設定されていないのにtextnextを呼びました。" end
		if @seq_text.size>@seq_no then
			@seq_f.put(@txt,@seq_text[@seq_no],@seq_x+@seq_no*@seq_f.getwidth,@seq_y)
			@seq_no=@seq_no+1
			if @seq_text.size>@seq_no then return false end
		end
		@seq_text=nil
		@seq_no=nil
		@seq_x=nil
		@seq_y=nil
		@seq_f=nil
		return true
	end
	def textstop(flushflag)
		#$wnd.textstop(flushflag)
		#@textstartで設定した文字列出力を終了します。
		#@flushflagにtrueを設定すると全部テキストを書き込んでから終了します。falseの場合は書き込まずにその場で終了します。
		if flushflag then
			while !textnext do	end
		end
		@seq_text=nil
		@seq_no=nil
		@seq_x=nil
		@seq_y=nil
		@seq_f=nil
	end
	def clear
		#$wnd.textclear
		#@テキストウィンドウをクリアします。
		@txt.fill(0)
	end
end

#!Selectクラス
#|ウィンドウからのパッド入力による選択肢の選択を処理するクラスです。
#|テキストをすべて表示し終わった後のWindowオブジェクトを指定し、このクラスのインスタンスを作成します。
#|このクラスのインスタンスを削除してもWindowが削除されるわけではありませんので、例えばマルチウィンドウの時に下に出したままに出来ます。
class Select
	def initialize (window,cur_array,cursor,row=nil)
		#$Select.new(window,[[curx,cury],[curx,cury],...],cursor[,row])→Select
		#@Selectオブジェクトを作成します。windowは選択肢に割り当てたいWindowオブジェクトです。
		#@[[curx,cury],...]は、一つ目の選択肢から順に、ウィンドウの左上からの座標を配列にした二重配列です。
		#@cursorにはカーソル画像としてAnimatorオブジェクトを渡します。
		#@rowに数値を指定すると複数列に対応します。一列の長さを指定してください（パッドの左右を押した時にそれだけ選択番号が進んだりもどったりします）
		@wnd=window
		@arr=cur_array
		@cur=cursor
		@sel=0
		@selnum=cur_array.size
		@row=row
		@counter=0
	end
	def draw (x,y,alpha=nil)
		#$sel.draw(x,y[,alpha])
		#@指定した位置に選択肢ウィンドウとカーソルを表示します。alphaは省略すると255になります。
		@wnd.draw(x,y,alpha)
		Draw.to(x+@arr[@sel][0],y+@arr[@sel][1])
		@cur.draw(alpha)
	end
	def pad
		#$sel.pad→Fixnum
		#@パッド入力による選択／決定処理をします。
		#@戻り値は、0以上の数は選択された選択肢番号、nilはまだ決定されていない状態、-1はキャンセルが押された状態です。
		if Input.getpad(0,Input::A)==2 then
			return @sel
		end
		if Input.getpad(0,Input::B)==2 then
			return -1
		end

		vx=Input.getpad(0,Input::VX)
		vy=Input.getpad(0,Input::VY)
		if vx==0 && vy==0 then
			@counter=0
		end
		if @counter>0 then
			@counter=@counter-1
			return nil
		end
		if vy!=0 then
			if vy<0 then
				@sel=@sel-1
				if @sel<0 then @sel=@arr.size-1 end
				@counter=15
			else
				@sel=@sel+1
				if @sel>=@arr.size then @sel=0 end
				@counter=15
			end
		elsif vx!=0 then
			if @row then
				if vx<0 then
					t=@sel-@row
					if t>=0 then
						@sel=t
					else
						@sel=@sel-1
						if @sel<0 then @sel=@arr.size-1 end
					end
					@counter=15
				else
					t=@sel+@row
					if t<@arr.size then
						@sel=t
					else
						@sel=@sel+1
						if @sel>=@arr.size then @sel=0 end
					end
					@counter=15
				end
			end
		end
		return nil
	end
end


#!デフォルトのゲームループの実装
def Game.initialize
	#$Game.initialize
	#@Fiberを作成しトップレベルのmainメソッドを呼んでいます。
	@fiber=Fiber.new do
		while true
			main
		end
	end
end

def Game.update
	#$Game.update
	#@ムービーテクスチャを更新し、シーンを描画して、mainメソッドに処理を戻します
	Sprite.update
	@fiber.resume
end

def Game.finalize
	#$Game.finalize
	#@デフォルトの実装では何もしていません。
end

#!トップレベルメソッド
#$doevents
#@フレームを描画し、メッセージを処理し、VSYNC管理をします。
#@メインルーチンで一フレーム分の処理が全部終わった時にこれを呼んでください。

def doevents
	Fiber.yield
	#現状これだけだけど、フレーム描画完了後の処理ということになるので、将来的に処理を増やすかも。
end